/*
 *  Copyright (C) 2002-20 - ntop.org
 *
 *  http://www.ntop.org/
 *
 * This code is proprietary code subject to the terms and conditions
 * defined in LICENSE file which is part of this source code package.
 */

#ifndef _DNS_PLUGIN_H_
#define _DNS_PLUGIN_H_ 1


#define QUERY_MAX_LEN        256 
#define RESPONSE_MAX_LEN     256

#define MAX_NUM_FILES      10000
#define TMP_FILE_POSTFIX   ".tmp"

/* Ports */
#define DNS_PORT             53
#define LLMNR_PORT           5355

PACKED_ON struct dns_packet_header{
  u_int16_t transaction_id, flags, num_queries, answer_rrs, authority_rrs /* , additional_rrs */;
} PACKED_OFF;

#define MAX_NUM_ANSWERS    32
#define MAX_NUM_AUTH_NS     8
#define MAX_ANSWER_LEN    128
#define MAX_AUTH_NS_LEN    82

struct a_answer {
  u_int32_t a_record, ttl;
};

typedef enum {
  unknown = 0,
  client2server,
  server2client
} DnsDirection;

struct str_dns_answer {
  u_int8_t answer_type;
  u_char answer[MAX_ANSWER_LEN];
};

struct dns_plugin_info {
  char dns_query[256];
  u_int16_t num_query_pkts, num_reply_pkts;
  u_int8_t flow_dumped;
  u_int8_t ret_code, num_answers, num_a_answers, num_auth_ns, query_class;
  u_int16_t transaction_id, query_type;
  u_int32_t first_rsp_ttl;
  u_char auth_ns[MAX_NUM_AUTH_NS][MAX_AUTH_NS_LEN];
  struct str_dns_answer answers[MAX_NUM_ANSWERS];
  struct a_answer a_answers[MAX_NUM_ANSWERS]; /* TODO: support AAAA (as string in .answers now) */
  DnsDirection src_to_dst_direction;
  /* DNS response */
  u_int8_t dns_response_computed, skip_flow, flow_lua_analyzed;
  struct timeval firstQueryTime, firstResponseTime;
  char dns_response_str[256], dns_response_buffer[4096];
  int32_t dns_response_buffer_len, last_tcpSeqNum;
};


#endif /* _DNS_PLUGIN_H_ */
